#include "LaunchProcess.h"
#include <stdio.h>
#include "global.h"
#include "RedirectInput.h"

// Function name	: Connect
// Description	    : 
// Return type		: bool 
// Argument         : LPWSTR host
// Argument         : IMPILaunch *&pLaunch
bool Connect(LPWSTR host, IMPILaunch *&pLaunch)
{
	HRESULT hr;
	MULTI_QI qi;
	WCHAR localhost[100];
	COSERVERINFO server;
	DWORD length = 100;

	qi.pIID = &IID_IMPILaunch;
	qi.pItf = NULL;

	server.dwReserved1 = 0;
	server.dwReserved2 = 0;
	server.pAuthInfo = NULL;
	server.pwszName = host;

	GetComputerNameW(localhost, &length);

	if (_wcsicmp(localhost, host) == 0)
	{
		hr = CoCreateInstanceEx(CLSID_MPILaunch, NULL, CLSCTX_SERVER, &server, 1, &qi);
		if (FAILED(hr))
		{
			wprintf(L"Unable to connect to %s: ", host);
			PrintError(hr);
			return false;
		}
	}
	else
	{
		hr = CoCreateInstanceEx(CLSID_MPILaunch, NULL, CLSCTX_REMOTE_SERVER, &server, 1, &qi);
		if (FAILED(hr))
		{
			wprintf(L"Unable to connect to %s: ", host);
			PrintError(hr);
			return false;
		}
	}

	if (SUCCEEDED(qi.hr))
	{
		pLaunch = (IMPILaunch*)qi.pItf;
		return true;
	}

	return false;
}

// Function name	: AbortThread
// Description	    : 
// Return type		: void 
// Argument         : IStream *pStream
void AbortThread(IStream *pStream)
{
	IMPILaunch *pLaunch=NULL;
	HRESULT hr=S_OK;
	long error=0;
	BSTR berror_msg;

	hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);

	CoGetInterfaceAndReleaseStream (pStream, IID_IMPILaunch, (void**) &pLaunch);

	WaitForSingleObject(g_hAbortEvent, INFINITE);

	if (g_bNormalExit)
	{
		pLaunch->Release();
		CoUninitialize();
		return;
	}
	
	error = 0;
	berror_msg = SysAllocString(L"");
	hr = pLaunch->Abort(&error, &berror_msg);
	if (FAILED(hr))
	{
		printf("Abort failed\n");
		PrintError(hr);
	}
	if (error)
	{
		if (wcslen(berror_msg) < 1)
			wprintf(L"PutInteractiveInput failed: %d\n", error);
		else
			wprintf(L"PutInteractiveInput failed: %s\n", berror_msg);
	}

	pLaunch->Release();
	CoUninitialize();
}

// Function name	: LaunchProcess
// Description	    : 
// Return type		: void 
// Argument         : LaunchProcessArg *arg
void LaunchProcess(LaunchProcessArg *arg)
{
	DWORD length = 100;
	RedirectInputThreadArg *rarg = NULL;
	HANDLE hRIThread = NULL;
	IMPILaunch *pLaunch;
	long error, pid;
	BSTR berror_msg;
	BSTR bFilename;
	BSTR bCmdLine, bEnv, bDir;
	HRESULT hr;
	DWORD dwThreadID;

	hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);
	if (FAILED(hr))
	{
		printf("CoInitializeEx failed: ");
		PrintError(hr);
		delete arg;
		return;
	}

	// Connect to MPILaunch
	if (Connect(arg->pszHost, pLaunch))
	{
		berror_msg = SysAllocString(L"");
		if (arg->i == 0)
		{
			bFilename = SysAllocString(L"");
			error = 0;
			hr = pLaunch->CreateTempFile(&bFilename, &error, &berror_msg);
			if (FAILED(hr))
			{
				wprintf(L"LaunchProcess:CreateTempFile failed on %s\n", arg->pszHost);
				PrintError(hr);
				pLaunch->Release();
				delete arg;
				SysFreeString(bFilename);
				ExitProcess(1);
			}
			if (error)
			{
				pLaunch->Release();
				wprintf(L"LaunchProces:%s", berror_msg);
				delete arg;
				SysFreeString(bFilename);
				ExitProcess(1);
			}
			wcscat(arg->pszEnv, L"|MPICH_EXTRA=");
			wcscat(arg->pszEnv, bFilename);
		}

		bCmdLine = SysAllocString(arg->pszCmdLine);
		bDir = SysAllocString(arg->pszDir);
		bEnv = SysAllocString(arg->pszEnv);
		berror_msg = SysAllocString(L"");

		//wprintf(L"Launching:\n exe: %s\n dir: %s\n env: %s\n", bCmdLine, bDir, bEnv);
		// LaunchProcess
		error = 0;
		if (arg->i == 0)
		{
			BSTR bNull = SysAllocString(L"");
			hr = pLaunch->GrantAccessToDesktop(bNull, bNull, &error, &berror_msg);
			SysFreeString(bNull);
		}
		if (arg->bLogon)
		{
			BSTR bAccount, bPassword;
			bAccount = SysAllocString(arg->pszAccount);
			bPassword = SysAllocString(arg->pszPassword);
			//if (arg->i == 0) pLaunch->GrantAccessToDesktop(bAccount, bPassword, &error, &berror_msg);
			//printf("LaunchInteractiveMPIProcess\n");fflush(stdout);
			hr = pLaunch->LaunchInteractiveMPIProcess(bCmdLine, bDir, bEnv, bAccount, bPassword, &pid, &error, &berror_msg);
			SysFreeString(bAccount);
			SysFreeString(bPassword);
		}
		else
			hr = pLaunch->LaunchInteractiveImpersontatedMPIProcess(bCmdLine, bDir, bEnv, &pid, &error, &berror_msg);
		SysFreeString(bCmdLine);
		SysFreeString(bDir);
		SysFreeString(bEnv);

		if (FAILED(hr))
		{
			printf("LaunchInteractiveImpersontatedMPIProcess failed.\n");
			PrintError(hr);
			pLaunch->Release();
			SysFreeString(berror_msg);
			CoUninitialize();
			if (arg->i == 0)
			{
				SysFreeString(bFilename);
				ExitProcess(1);
			}
			delete arg;
			return;
		}
		if (error)
		{
			if (wcslen(berror_msg) < 1)
				wprintf(L"LaunchProcess:LaunchInteractiveImpersonatedMPIProcess failed: Error %d\n", error);
			else
				wprintf(L"LaunchProcess:%s\n", berror_msg);
			pLaunch->Release();
			CoUninitialize();
			SysFreeString(berror_msg);
			if (arg->i == 0)
			{
				SysFreeString(bFilename);
				ExitProcess(1);
			}
			delete arg;
			return;
		}
		
		// Get the port number and redirect input to the first process
		if (arg->i == 0)
		{
			error = 0;
			hr = pLaunch->GetPortFromFile(bFilename, &g_nRootPort, &error, &berror_msg);
			SysFreeString(bFilename);
			if (FAILED(hr))
			{
				pLaunch->Release();
				printf("LaunchProcess:GetPortFromFile failed\n");
				delete arg;
				ExitProcess(1);
			}
			if (error)
			{
				pLaunch->Release();
				wprintf(L"LaunchProcess:%s", berror_msg);
				delete arg;
				ExitProcess(1);
			}

			rarg = new RedirectInputThreadArg;
			rarg->hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

			rarg->ppStream = new IStream*;
			// Marshall pLaunch to a thread which redirects user input
			hr = CoMarshalInterThreadInterfaceInStream(IID_IMPILaunch, pLaunch, rarg->ppStream);
			if (FAILED(hr))
			{
				pLaunch->Release();
				SysFreeString(berror_msg);
				printf("LaunchProcess:CoMarshalInterThreadInterfaceInStream failed.\n");
				PrintError(hr);
				CoUninitialize();
				delete rarg;
				delete arg;
				ExitProcess(1);
			}
			
			hRIThread = CreateThread(
				NULL, 0, 
				(LPTHREAD_START_ROUTINE)RedirectInputThread,
				rarg, 
				0, &dwThreadID);
			
			if (hRIThread == NULL)
			{
				error = GetLastError();
				printf("CreateThread failed: error %d\n", error);
				PrintError(error);
				pLaunch->Release();
				SysFreeString(berror_msg);
				CoUninitialize();
				delete rarg;
				delete arg;
				ExitProcess(1);
			}
		}
		
		IStream *pStream;
		HANDLE hAbortThread = NULL;
		// Marshall pLaunch to a thread which waits for an abort event
		hr = CoMarshalInterThreadInterfaceInStream(IID_IMPILaunch, pLaunch, &pStream);
		if (FAILED(hr))
		{
			pLaunch->Release();
			SysFreeString(berror_msg);
			printf("LaunchProcess:CoMarshalInterThreadInterfaceInStream failed.\n");
			PrintError(hr);
			CoUninitialize();
			delete rarg;
			delete arg;
			ExitProcess(1);
		}
		
		hAbortThread = CreateThread(
			NULL, 0, 
			(LPTHREAD_START_ROUTINE)AbortThread,
			pStream, 
			0, &dwThreadID);
		
		if (hAbortThread == NULL)
		{
			error = GetLastError();
			printf("CreateThread failed: error %d\n", error);
			PrintError(error);
			pLaunch->Release();
			SysFreeString(berror_msg);
			CoUninitialize();
			delete rarg;
			delete arg;
			ExitProcess(1);
		}
		CloseHandle(hAbortThread);
		
		// Redirect output
		HANDLE hStdout = GetStdHandle(STD_OUTPUT_HANDLE);

		if (hStdout == INVALID_HANDLE_VALUE)
		{
			error = GetLastError();
			_tprintf(TEXT("GetStdHandle failed: Error %d\n"), error);
			pLaunch->Release();
			CoUninitialize();
			SysFreeString(berror_msg);
			if (arg->i == 0) SetEvent(rarg->hEvent);
			delete arg;
			return;
		}

		long more = 1, from = 0;
		VARIANT *v = new VARIANT;
		DWORD num_written, num_elements;
		void *pBuf;
		
		v->vt = VT_UI1 | VT_ARRAY;
		SAFEARRAYBOUND bound;
		bound.lLbound = 0;
		bound.cElements = 0;
		v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
		
		while (more == 1)
		{
			error = 0;
			hr = pLaunch->GetInteractiveOutput(v, &more, &error, &berror_msg);
			if (FAILED(hr))
			{
				_tprintf(TEXT("LaunchProcess:GetInteractiveOutput failed\n"));
				PrintError(hr);
				SysFreeString(berror_msg);
				VariantClear(v);
				delete v;
				if (arg->i == 0) SetEvent(rarg->hEvent);
				delete arg;
				return;
			}
			if (error)
			{
				if (wcslen(berror_msg) < 1)
					wprintf(L"LaunchProcess:GetInteractiveOutput:Error %d", error);
				else
					wprintf(L"LaunchProcess:%s", berror_msg);
				SysFreeString(berror_msg);
				VariantClear(v);
				delete v;
				if (arg->i == 0) SetEvent(rarg->hEvent);
				delete arg;
				return;
			}
			if (v->parray != NULL)
			{
				num_elements = v->parray->rgsabound->cElements;
				if (num_elements > 0)
				{
					SafeArrayAccessData(v->parray, &pBuf);
					WaitForSingleObject(g_hConsoleOutputMutex, 5000);
#ifdef MULTI_COLOR_OUTPUT
					SetConsoleTextAttribute(hStdout, aConsoleColorAttribute[arg->i%NUM_OUTPUT_COLORS]);
					WriteFile(hStdout, pBuf, num_elements, &num_written, NULL);
					SetConsoleTextAttribute(hStdout, g_ConsoleAttribute);
#else
					WriteFile(hStdout, pBuf, num_elements, &num_written, NULL);
#endif
					FlushFileBuffers(hStdout);
					ReleaseMutex(g_hConsoleOutputMutex);
					SafeArrayUnaccessData(v->parray);
					// Destroy the array.
					SafeArrayDestroy(v->parray);
					// Create a new and empty array.
					v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
				}
				else
				{
					SafeArrayDestroy(v->parray);
					v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
				}
			}
			else
			{
				v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
			}
		}
		
		SafeArrayDestroy(v->parray);
		delete v;

		// Stop the redirect input thread
		if (arg->i == 0)
		{
			SetEvent(rarg->hEvent);
			WaitForSingleObject(hRIThread, 5000);
			CloseHandle(hRIThread);
		}
	}
	else
	{
		wprintf(L"Connect to %s failed\n", arg->pszHost);
		if (arg->i == 0)
			ExitProcess(1);
	}
		
	delete arg;
}
