#ifndef BNR_H
#define BNR_H

#if defined(__cplusplus)
extern "C" {
#endif

#ifdef MPICHBNR_EXPORTS
#define MPICHBNR_API __declspec(dllexport)
#else
#define MPICHBNR_API __declspec(dllimport)
#endif

/*
struct MPIR_Info {
    int cookie;
    char *key, *value;
    struct MPIR_Info *next;
};
//*/
typedef struct MPIR_Info *MPI_Info;
typedef int BNR_gid;

#define BNR_SUCCESS		0
#define BNR_FAIL		1
#define BNR_MAXATTRLEN	64
#define BNR_MAXVALLEN	3*1024

/* returns primary group id assigned at creation */
MPICHBNR_API int BNR_Init( BNR_gid *mygid );

/* frees any internal resources
 * No BNR calls may be made after BNR_Cleanup
 */
MPICHBNR_API int BNR_Cleanup();

/* barriers all processes in group; puts done
 * before the fence are accessible by gets after
 * the fence 
 */
MPICHBNR_API int BNR_Fence( BNR_gid );	   

/* puts attr-value pair for retrieval by other
 * processes in group;  attr is a string of
 * length < BNR_MAXATTRLEN, val is string of
 * length < BNR_MAXVALLEN 
 */
MPICHBNR_API int BNR_Put( BNR_gid group, char *attr, char *val );

/* matches attr, retrieves corresponding value
 * into val, which is a buffer of
 * length = BNR_MAXVALLEN 
 */
MPICHBNR_API int BNR_Get( BNR_gid group, char *attr, char *val );

/* returns rank in group */
MPICHBNR_API int BNR_Rank( BNR_gid group, int *myrank );

/* allocates a new, unique group id */
MPICHBNR_API int BNR_Allocate_gid( BNR_gid *new_gid );

/* collective over group, arguments valid only at
 * root.  Note we pass *in* the new group id,
 * assumed to have been BNR_allocated.  notify_fn
 * is called if a process exits, and gets the
 * group, rank, and return code. argv and env
 * arrays are null terminated.  remote_group_parent
 * is what the new processes will get as parent 
 */
MPICHBNR_API int BNR_Spawn( BNR_gid group, int root, BNR_gid remote_group, BNR_gid temp_gid, 
						   BNR_gid remote_group_parent, char *command, char *argv[], char *env[], 
						   MPI_Info info, int array_of_errcodes[], int (notify_fn)(BNR_gid group, 
						   int rank, int exit_code) );

/* calling process must be in the local group and
 * must not be in the remote group.  Collective 
 * over the union of the two groups. 
 */
MPICHBNR_API int BNR_Merge( BNR_gid local_group, BNR_gid remote_group, BNR_gid *new_group );

/* like BNR_Spawn, with arrays of length count */
MPICHBNR_API int BNR_Spawn_multiple( BNR_gid group, int root, BNR_gid remote_group,
                        BNR_gid remote_group_parent, int count,
						char *array_of_commands[], char *array_of_argv[],
						char *array_of_env[], MPI_Info array_of_info[],
						int array_of_errcodes[],
						int (notify_fn)(BNR_gid group, int rank, int exit_code) );

/* returns 1 if no parent all */
MPICHBNR_API int BNR_Parent( BNR_gid *parent_gid );

/* frees gid for re-use. */
MPICHBNR_API int BNR_Free_gid( BNR_gid gid );

/* kills processes in group given by gid.  This
 * can be used, for example, during spawn_multiple
 * when a spawn fails, to kill off groups already
 * spawned before returning failure 
 */
MPICHBNR_API int BNR_Kill( BNR_gid gid );

/* The following are needed for publishing.  They require no fence, since they are not
 * assumed to be either scalable or local.  The inevitable race condition is just
 * accepted.
 */ 

/* deposits attr-value pair for access */
MPICHBNR_API int BNR_Deposit( char *attr, char *value );

/* withdraws attr-value pair */
MPICHBNR_API int BNR_Withdraw( char *attr, char *value );

/* finds value of attribute */
MPICHBNR_API int BNR_Lookup( char *attr, char *value );

#if defined(__cplusplus)
}
#endif 

#endif
